(*
Author:  Christian Sternagel <c.sternagel@gmail.com> (2016)
License: LGPL (see file COPYING.LESSER)
*)

section \<open>Retrogression\<close>

theory Retrogression
imports
  Peak_Decreasingness
  "../Auxiliaries/Multiset_Extension2"
  "../Orderings/Lexicographic_Extension"
  Abstract_Completion
begin

locale ars_retrogressive =
  ars_labeled_sn +
  fixes J
  assumes J_subset: "J \<subseteq> I"
    and retro: "\<forall>a \<in> I - J. r a \<subseteq> (UNION {b\<in>I. a \<succ> b} r)\<^sup>\<leftrightarrow>\<^sup>*"
begin


abbreviation "S \<equiv> (\<Union>i\<in>J. r i)"

lemma conversion_equiv:
  "R\<^sup>\<leftrightarrow>\<^sup>* = S\<^sup>\<leftrightarrow>\<^sup>*"
proof -
  let ?M = "s_mul_ext Id ({\<succ>}\<^sup>+)"
  interpret SN_order_pair "{\<succ>}\<^sup>+" Id
    by (standard) (insert SN_less, auto simp: refl_on_def trans_def SN_trancl_SN_conv)
  interpret SN_order_pair "?M" "ns_mul_ext Id ({\<succ>}\<^sup>+)"
    by (intro SN_order_pair.mul_ext_SN_order_pair) fact

  { fix x y assume "(x, y) \<in> R\<^sup>\<leftrightarrow>\<^sup>*"
    then obtain M where "conv M x y" by (auto simp: conversion_conv_iff)
    then have "(x, y) \<in> S\<^sup>\<leftrightarrow>\<^sup>*"
    proof (induct M arbitrary: x y rule: SN_induct [OF SN])
      case (1 M)
      note IH = 1(1)
      show ?case using \<open>conv M x y\<close>
      proof (cases)
        case (step a z N)
        then have "(M, N) \<in> ?M" by (intro s_mul_ext_IdI [of "{#a#}" _ N _ "{#}"]) auto
        from IH [OF this \<open>conv N z y\<close>] have "(z, y) \<in> S\<^sup>\<leftrightarrow>\<^sup>*" .
        moreover have "(x, z) \<in> S\<^sup>\<leftrightarrow>\<^sup>*" using step
        proof (cases "a \<in> J")
          case False
          with step and retro have "(x, z) \<in> (\<Union>b \<in> (a)\<succ>. r b)\<^sup>\<leftrightarrow>\<^sup>*"
            by auto (meson DiffI conversion_inv subsetCE)
          then obtain K where "set_mset K \<subseteq> (a)\<succ>" and "conv K x z"
            by (auto simp: conversion_conv_iff [OF Collect_restrict])
          then have "(M, K) \<in> ?M" by (intro s_mul_ext_IdI [of M _ "{#}" _ K]) (auto simp: step)
          from IH [OF this \<open>conv K x z\<close>] show ?thesis .
        qed blast
        ultimately show ?thesis by (auto simp: conversion_def)
      qed simp
    qed
  }
  moreover
  { fix x y assume "(x, y) \<in> S\<^sup>\<leftrightarrow>\<^sup>*"
    then have "(x, y) \<in> R\<^sup>\<leftrightarrow>\<^sup>*"
      using J_subset by (auto simp: conversion_def intro: rtrancl_mono [THEN subsetD, of "S\<^sup>\<leftrightarrow>" "R\<^sup>\<leftrightarrow>"])
  }
  ultimately show ?thesis by auto
qed

end

context kb_irun
begin

lemma
  "(mstep M ((\<Union>i. E i) \<union> (\<Union>i. R i)))\<^sup>\<leftrightarrow>\<^sup>* = (mstep M (E\<^sub>\<omega> \<union> R\<^sub>\<omega>))\<^sup>\<leftrightarrow>\<^sup>*"
  (is "(mstep M ?I)\<^sup>\<leftrightarrow>\<^sup>* = (mstep M ?J)\<^sup>\<leftrightarrow>\<^sup>*")
proof -
  let ?m2 = "case_prod (\<lambda>x y. {#x, y#})"
  def tpless (infix "\<sqsupset>" 50) \<equiv> "\<lambda>x y. (?m2 x, ?m2 y) \<in> mulless"
  have "SN {(x, y). x \<sqsupset> y}" using mlessop.SN by (auto simp: SN_defs tpless_def)
  interpret ars_retrogressive "\<lambda>a. mstep M {a}" ?I tpless ?J
  proof
    note [dest!] = R_per_subset_R_inf [THEN subsetD] E_per_subset_E_inf [THEN subsetD]
    show "SN {(x, y). x \<sqsupset> y}" by fact
    show "?J \<subseteq> ?I" by blast
    show "\<forall>a\<in>?I - ?J. mstep M {a} \<subseteq> (\<Union>a\<in>{b \<in> ?I. a \<sqsupset> b}. mstep M {a})\<^sup>\<leftrightarrow>\<^sup>*"
    proof (intro ballI subrelI)
      fix a s t assume "a \<in> ?I - ?J" and "(s, t) \<in> mstep M {a}"
      then consider "a \<in> (\<Union>i. E i) - (E\<^sub>\<omega> \<union> (\<Union>i. R i))" | "a \<in> (\<Union>i. R i) - R\<^sub>\<omega>" by auto
      then show "(s, t) \<in> (\<Union>a\<in>{b \<in> ?I. a \<sqsupset> b}. mstep M {a})\<^sup>\<leftrightarrow>\<^sup>*"
      proof (cases)
        case 1 then show ?thesis sorry
      next
        case 2 then show ?thesis sorry
      qed
    qed
  qed
  have *: "(\<Union>a\<in>?I. mstep M {a}) = mstep M ?I" "(\<Union>a\<in>?J. mstep M {a}) = mstep M ?J"
    by (simp_all add: UN_mstep)
  show ?thesis using conversion_equiv unfolding * by simp
qed

end

context kb
begin

definition "lstep' E R M N = {(s, t). \<exists>l r s' t'. s' \<in># M \<and> t' \<in># M \<and> s' \<succeq> s \<and> t' \<succeq> t \<and>
  (N = {#l, l, r, r#} \<and> (l, r) \<in> E \<and> (s, t) \<in> rstep {(l, r)} \<or>
   N = {#l, r#} \<and> (l, r) \<in> R \<and> (s, t) \<in> rstep {(l, r)})}"

lemma lstep'_cases:
  assumes "(s, t) \<in> lstep' E R M N"
  obtains l r s' t' where "s' \<in># M" and "t' \<in># M" and "s' \<succeq> s" and "t' \<succeq> t"
    and "N = {#l, l, r, r#}" and "(l, r) \<in> E" and "(s, t) \<in> rstep {(l, r)}"
  | l r s' t' where "s' \<in># M" and "t' \<in># M" and "s' \<succeq> s" and "t' \<succeq> t"
    and "N = {#l, r#}" and "(l, r) \<in> R" and "(s, t) \<in> rstep {(l, r)}"
using assms by (auto simp: lstep'_def)

lemma lstep'I:
  assumes "s' \<in>#M" and "t' \<in># M" and "s' \<succeq> s" and "t' \<succeq> t"
    and "(s, t) \<in> rstep {(l, r)}"
    and "(N = {#l, l, r, r#} \<and> (l, r) \<in> E) \<or> (N = {#l, r#} \<and> (l, r) \<in> R)"
  shows "(s, t) \<in> lstep' E R M N"
using assms by (auto simp: lstep'_def)

lemma lstep'_E_I:
  "(l, r) \<in> E \<Longrightarrow> s' \<in># M \<Longrightarrow> t' \<in>#M \<Longrightarrow> s' \<succeq> s \<Longrightarrow> t' \<succeq> t \<Longrightarrow> (s, t) \<in> rstep {(l, r)} \<Longrightarrow>
    (s, t) \<in> lstep' E R M {#l, l, r, r#}"
by (auto simp: lstep'_def)

lemma lstep'_R_I:
  "(l, r) \<in> R \<Longrightarrow> s' \<in># M \<Longrightarrow> t' \<in>#M \<Longrightarrow> s' \<succeq> s \<Longrightarrow> t' \<succeq> t \<Longrightarrow> (s, t) \<in> rstep {(l, r)} \<Longrightarrow>
    (s, t) \<in> lstep' E R M {#l, r#}"
by (auto simp: lstep'_def)

end

context kb_irun
begin

lemma
  "(mstep M ((\<Union>i. E i) \<union> (\<Union>i. R i)))\<^sup>\<leftrightarrow>\<^sup>* = (mstep M (E\<^sub>\<omega> \<union> R\<^sub>\<omega>))\<^sup>\<leftrightarrow>\<^sup>*"
proof -
  let ?E = "\<Union>i. E i"
  let ?R = "\<Union>i. R i"
  let ?E2 = "\<lambda>E. {{#l, l, r, r#} | l r. (l, r) \<in> E}"
  let ?R1 = "\<lambda>R. {{#l, r#} | l r. (l, r) \<in> R}"
  def I \<equiv> "?E2 ?E \<union> ?R1 ?R"
  def J \<equiv> "?E2 E\<^sub>\<omega> \<union> ?R1 R\<^sub>\<omega>"
  note [dest!] = R_per_subset_R_inf [THEN subsetD] E_per_subset_E_inf [THEN subsetD]
  have [intro!]: "\<And>i l r. (l, r) \<in> E i \<Longrightarrow> {#l, l, r, r#} \<in> I" by (auto simp: I_def)
  have [intro!]: "\<And>i l r. (l, r) \<in> R i \<Longrightarrow> {#l, r#} \<in> I" by (auto simp: I_def)
  have [simp]: "(\<Union>a\<in>I. lstep' ?E ?R M a) = mstep M (?E \<union> ?R)"
    apply (auto elim!: lstep'_cases simp: mstep_iff)
    using gr_implies_not0
    apply (blast intro: lstep'_E_I lstep'_R_I)+
    done
  have [simp]: "(\<Union>a\<in>J. lstep' ?E ?R M a) = mstep M (E\<^sub>\<omega> \<union> R\<^sub>\<omega>)"
    apply (auto elim!: lstep'_cases simp: mstep_iff)
    using gr_implies_not0
    (*apply (blast intro: lstep'_E_I lstep'_R_I)+*)
   sorry
  interpret ars_retrogressive "lstep' ?E ?R M" I "\<lambda>x y. (x, y) \<in> mulless" J
  proof
    show "SN {(x, y). (x, y) \<in> mulless}" by (auto intro: SN_mulless)
    show "J \<subseteq> I" by (auto simp: J_def I_def)
    show "\<forall>a\<in>I - J.
       lstep' (\<Union>i. E i) (\<Union>i. R i) M a
       \<subseteq> (UNION {b \<in> I. (a, b) \<in> mulless}
            (lstep' (\<Union>i. E i) (\<Union>i. R i) M))\<^sup>\<leftrightarrow>\<^sup>*" sorry
  qed
  show ?thesis using conversion_equiv by simp
qed

end

end