theory Foo
imports Trs
begin

definition "GT1 R = {(f, g). \<exists>(l, r)\<in>R. root l = f \<and> (root r = g \<or> root r = None)}"

abbreviation "GE R \<equiv> (GT1 R)\<^sup>*"
abbreviation "GT R \<equiv> (GT1 R)\<^sup>+"

lemma collapsing_imp_GT1:
  assumes "(Fun f ts, Var x) \<in> R"
  shows "(Some (f, length ts), y) \<in> GT1 R"
using assms by (force simp: GT1_def)

lemma rrstep_GT1:
  assumes "(s, t) \<in> rrstep R" and var: "\<forall>(l, r)\<in>R. is_Fun l"
  shows "(root s, root t) \<in> GT1 R"
using assms(1)
proof (rule rrstepE)
  fix \<sigma> l r assume "(l, r) \<in> R" and "s = l \<cdot> \<sigma>" and "t = r \<cdot> \<sigma>"
  then show ?thesis
    using var [THEN bspec, OF \<open>(l, r) \<in> R\<close>]
    by (cases l; cases r) (force dest: collapsing_imp_GT1 simp: GT1_def)+
qed

lemma rstep_GE:
  assumes "(s, t) \<in> rstep R" and var: "\<forall>(l, r)\<in>R. is_Fun l"
  shows "(root s, root t) \<in> GE R"
using assms
apply (cases rule: rstep_cases)
apply (auto dest: rrstep_GT1 elim!: nrrstepE)
apply (case_tac C)
apply auto
done

lemma rsteps_GE:
  assumes "(s, t) \<in> (rstep R)\<^sup>*" and "\<forall>(l, r)\<in>R. is_Fun l"
  shows "(root s, root t) \<in> GE R"
using assms by (induct) (auto dest: rstep_GE)

inductive rrsteps_n for R
where
  nrrsteps: "(s, t) \<in> (nrrstep R)\<^sup>* \<Longrightarrow> rrsteps_n R 0 s t"
| rrstep: "(s, t) \<in> rrstep R \<Longrightarrow> rrsteps_n R (Suc 0) s t"
| trans: "rrsteps_n R m s t \<Longrightarrow> rrsteps_n R n t u \<Longrightarrow> rrsteps_n R (m + n) s u"

lemma rrsteps_n_refl [intro, simp]: "rrsteps_n R 0 t t" by (auto intro: rrsteps_n.intros)

lemma rsteps_imp_rrsteps_n:
  assumes "(s, t) \<in> (rstep R)\<^sup>*"
  shows "\<exists>n. rrsteps_n R n s t"
using assms
proof (induct)
  case (step t u)
  then show ?case
    by (auto elim: rstep_cases dest: rrsteps_n.trans [OF _ rrstep] rrsteps_n.trans [OF _ nrrsteps])
qed auto

lemma rrsteps_n_imp_rsteps:
  assumes "rrsteps_n R n s t"
  shows "(s, t) \<in> (rstep R)\<^sup>*"
using assms by (induct) (auto dest: rrstep_imp_rstep nrrsteps_imp_rsteps)

lemma rrsteps_n_0:
  assumes "rrsteps_n R 0 s t"
  shows "root s = root t \<and> (s, t) \<in> (nrrstep R)\<^sup>*"
using assms by (induct "0::nat" s t) (auto simp: nrrsteps_imp_eq_root_arg_rsteps)

lemma rrsteps_n_Suc:
  assumes "rrsteps_n R (Suc n) s t" and "\<forall>(l, r)\<in>R. is_Fun l"
  shows "(root s, root t) \<in> GT R"
using assms
apply (induct "Suc n" s t arbitrary: n)
apply (auto dest: rrstep_GT1)
apply (case_tac m; case_tac n)
apply (auto dest: rrsteps_n_0 trancl_trans)
done

lemma not_GT_rsteps_iff:
  assumes "(Some (f, length (ss @ ts) + 1), Some (f, length (ss @ ts) + 1)) \<notin> GT R"
    and "\<forall>(l, r)\<in>R. is_Fun l"
  shows "(Fun f (ss @ t # ts), Fun f (ss @ u # ts)) \<in> (rstep R)\<^sup>* \<longleftrightarrow> (t, u) \<in> (rstep R)\<^sup>*"
    (is "(?C t, ?C u) \<in> _ \<longleftrightarrow> _")
proof
  assume "(?C t, ?C u) \<in> (rstep R)\<^sup>*"
  then obtain n where "rrsteps_n R n (?C t) (?C u)" by (auto dest: rsteps_imp_rrsteps_n)
  then show "(t, u) \<in> (rstep R)\<^sup>*"
    using assms
    by (cases n)
       (auto dest!: rrsteps_n_Suc rrsteps_n_0 nrrsteps_imp_arg_rsteps [where i = "length ss"])
qed (auto dest: rsteps_closed_ctxt [where C = "More f ss \<box> ts"])

lemma not_GT_rsteps_iff':
  assumes "(Some (f, length ts), Some (f, length ts)) \<notin> GT R"
    and "length ss = length ts"
    and "\<forall>(l, r)\<in>R. is_Fun l"
  shows "(Fun f ss, Fun f ts) \<in> (rstep R)\<^sup>* \<longleftrightarrow> (\<forall>i<length ss. (ss ! i, ts ! i) \<in> (rstep R)\<^sup>*)"
proof
  assume "(Fun f ss, Fun f ts) \<in> (rstep R)\<^sup>*"
  then obtain n where "rrsteps_n R n (Fun f ss) (Fun f ts)" by (auto dest: rsteps_imp_rrsteps_n)
  then show "\<forall>i<length ss. (ss ! i, ts ! i) \<in> (rstep R)\<^sup>*"
    using assms
    by (cases n) (auto dest!: rrsteps_n_Suc rrsteps_n_0 nrrsteps_imp_arg_rsteps)
qed (rule args_rsteps_imp_rsteps [OF assms(2)])

context
  fixes R :: "('f, 'v) trs"
begin

declare ex_cong [fundef_cong]

fun nonreach
where
"nonreach (Fun f ss) (Fun g ts) =
    (if (Some (f, length ss), Some (g, length ts)) \<notin> GE R then True
    else if f = g \<and> length ss = length ts \<and> (Some (g, length ts), Some (g, length ts)) \<notin> GT R then
      (\<exists>i<length ts. nonreach (ss ! i) (ts ! i))
    else False)"
| "nonreach s t = False"

fun nonreach'
where
  "nonreach' (Fun f ss) (Fun g ts) \<longleftrightarrow>
    (Some (f, length ss), Some (g, length ts)) \<notin> GT R \<and>
    (f = g \<and> length ss = length ts \<longrightarrow> (\<exists>i<length ts. nonreach' (ss ! i) (ts ! i)))"
| "nonreach' s t = False"

lemma
  "nonreach' s t = nonreach s t"
by (induct s t rule: nonreach'.induct) (auto simp: rtrancl_eq_or_trancl)

fun reachs
where
  "reachs (Fun f ss, Fun g ts) =
    (if f = g \<and> length ss = length ts \<and> (Some (g, length ts), Some (g, length ts)) \<notin> GT R then
      concat (map reachs (zip ss ts))
    else [(Fun f ss, Fun g ts)])"
| "reachs (s, t) = [(s, t)]"

fun reachs' and reachs_list'
where
  "reachs' (Fun f ss) (Fun g ts) =
    (if f = g \<and> length ss = length ts \<and> (Some (g, length ts), Some (g, length ts)) \<notin> GT R then
      reachs_list' ss ts
    else [(Fun f ss, Fun g ts)])"
| "reachs' s t = [(s, t)]"
| "reachs_list' [] ts = []"
| "reachs_list' ss [] = []"
| "reachs_list' (s#ss) (t#ts) = reachs' s t @ reachs_list' ss ts"

lemma
  fixes s t :: "('f, 'v) term" and ss ts :: "('f, 'v) term list"
  shows "reachs' s t = reachs (s, t)" and "reachs_list' ss ts = concat (map reachs (zip ss ts))"
by (induct s t and ss ts rule: reachs'_reachs_list'.induct) simp_all

end


lemma GT_empty [simp]: "GT {} = {}" by (auto simp: GT1_def dest: tranclD)

lemma foo:
  assumes var: "\<forall>(l, r)\<in>R. is_Fun l"
  shows "(s, t) \<in> (rstep R)\<^sup>* \<Longrightarrow> \<forall>(u, v) \<in> set (reachs' R s t). (u, v) \<in> (rstep R)\<^sup>*"
    and "length ss = length ts \<Longrightarrow> \<forall>i<length ts. (ss ! i, ts ! i) \<in> (rstep R)\<^sup>* \<Longrightarrow>
      \<forall>(u, v) \<in> set (reachs_list' R ss ts). (u, v) \<in> (rstep R)\<^sup>*"
apply (induct s t and ss ts taking: "{}" rule: reachs'_reachs_list'.induct)
defer
apply simp
apply simp
apply simp
apply simp
defer
apply simp
apply (auto simp: all_set_conv_all_nth dest!: not_GT_rsteps_iff' [OF _ _ var])[1]
apply (simp add: all_set_conv_all_nth del: set_append)
apply auto
apply (case_tac "i < length (reachs' R s t)")
apply (auto simp: nth_append)
by (metis (no_types, hide_lams) Suc_less_eq2 add.commute add_diff_cancel_right' add_diff_inverse_nat less_Suc_eq less_diff_conv less_irrefl_nat linorder_neqE_nat nat_add_left_cancel_less not_less_eq nth_Cons_Suc)

lemma
  assumes var: "\<forall>(l, r)\<in>R. is_Fun l"
    and "nonreach R s t"
  shows "(s, t) \<notin> (rstep R)\<^sup>*"
using assms(2)
apply (induct s t taking: "{}" rule: nonreach.induct)
apply (auto split: if_splits dest: rsteps_GE [OF _ var] dest!: not_GT_rsteps_iff' [OF _ _ var])
done

lemma nonreach:
  assumes var: "\<forall>(l, r)\<in>R. is_Fun l"
    and "nonreach R s t"
  shows "\<not> (\<exists>\<sigma> \<tau>. (s \<cdot> \<sigma>, t \<cdot> \<tau>) \<in> (rstep R)\<^sup>*)"
using assms(2)
apply (induct s t taking: "{}" rule: nonreach.induct)
apply (auto split: if_splits dest: rsteps_GE [OF _ var])
using not_GT_rsteps_iff' [OF _ _ var]
apply (auto)
done

end
